//
//  MYCoroutine.m
//  Coroutines
//
//  Created by Jens Alfke on 4/29/08.
//  Copyright 2008 Jens Alfke. All rights reserved.
//  License is at the bottom of this file.
//

#import "MYCoroutine.h"
#import "CoroX.h"


#ifndef LogTo
#define kEnableLog 0    /* 1 enables logging, 0 disables it */
#define LogTo(DOMAIN,MSG,...) do{ if(kEnableLog) NSLog(@""#DOMAIN ": " MSG,__VA_ARGS__); }while(0)
#endif

#ifndef Warn
#define Warn(MSG,...) NSLog(@"WARNING: " #MSG,__VA_ARGS__)
#endif


@implementation MYCoroutine


static MYCoroutine *sMain, *sCurrent;


+ (void) initialize
{
    if( self == [MYCoroutine class] ) {
        sMain = [[self alloc] init];
        Coro_initializeMainCoro(sMain->_coro);
        sMain.name = @"MAIN";
        sCurrent = sMain;
    }
}


- (id) init
{
    self = [super init];
    if (self != nil) {
        _coro = Coro_new();
        LogTo(Coroutine,@"INIT %@ : _coro=%p",self,_coro);
    }
    return self;
}


- (void) dealloc
{
    LogTo(Coroutine,@"DEALLOC %@",self);
    Coro_free(_coro);
    [super dealloc];
}


- (NSString*) description
{
    if( _name )
        return [NSString stringWithFormat: @"%@[%@]", [self class],_name];
    else
        return [NSString stringWithFormat: @"%@[%p]", [self class],self];
}


@synthesize name=_name;


+ (MYCoroutine*) mainCoroutine          {return sMain;}
+ (MYCoroutine*) currentCoroutine       {return sCurrent;}

- (BOOL) isCurrent                      {return self==sCurrent;}

- (const void*) stack                   {return Coro_stack(_coro);}
- (size_t) stackSize                    {return Coro_stackSize(_coro);}
- (void) setStackSize: (size_t)size     {Coro_setStackSize_(_coro,size);}

- (size_t) bytesLeftOnStack             {return Coro_bytesLeftOnStack(_coro);}
- (BOOL) stackSpaceAlmostGone           {return Coro_stackSpaceAlmostGone(_coro);}


static void startWithInvocation( void *context )
{
    [(NSInvocation*)context invoke];
}

static void startWithMain( void *context )
{
    [(MYCoroutine*)context main];
}

- (void) startWithInvocation: (NSInvocation*)invocation
{
    LogTo(Coroutine,@"Starting %@ (currently in %@)",self,sCurrent);
    MYCoroutine *current = sCurrent;
    sCurrent = self;
    
    if( invocation )
        Coro_startCoro_(current->_coro, _coro, invocation, &startWithInvocation);
    else
        Coro_startCoro_(current->_coro, _coro, self, &startWithMain);
    
    sCurrent = current;
    LogTo(Coroutine,@"...resumed %@ after starting %@",sCurrent,self);
}

- (void) start
{
    [self startWithInvocation: nil];
}


+ (MYCoroutine*) startWithInvocation: (NSInvocation*)invocation
{
    MYCoroutine *cr = [[self alloc] init];
    [cr startWithInvocation: invocation];
    return cr;
}


- (void) resume
{
    LogTo(Coroutine,@"Resuming %@ (currently in %@)",self,sCurrent);
    MYCoroutine *current = sCurrent;
    sCurrent = self;
    Coro_switchTo_(current->_coro,_coro);
    sCurrent = current;
    LogTo(Coroutine,@"...resumed %@",sCurrent);
}


- (void) main
{
    // subclasses should override this.
}


@end




/*
 (This is a BSD License)
 
 Copyright (c) 2008 Jens Alfke
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without modification, are
 permitted provided that the following conditions are met:
 
 • Redistributions of source code must retain the above copyright notice, this list of
 conditions and the following disclaimer.
 • Redistributions in binary form must reproduce the above copyright notice, this list
 of conditions and the following disclaimer in the documentation and/or other materials
 provided with the distribution.
 • Neither the name of the author nor the names of other contributors may be used to
 endorse or promote products derived from this software without specific prior written
 permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
 EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
